//
// Generic benchmark program
// needs to be linked against actual Dbi-implementation
//
// Author: Christoph Dalitz, June 2004
//

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

#include "dbi.h"

// number of hersteller
#define NUM_HERSTELLER 100
// number of products per hersteller
#define NUM_PRODUKT    100
// number of lieferungen per produkt
#define NUM_LIEFERUNG  10

// number of tuples before commit in long transactions
#define NUM_COMMIT     10000

// set to 1 if only inserts are to be tested
#define INSERT_ONLY 0

// command line arguments
struct c_opt {
  c_opt() {
    dbname = "";
    user = "";
    password = "";
    verbose = 0;
  }
  string dbname;
  string user;
  string password;
  int verbose;
} opt;


//--------------------------------------------------------------------
// function for time measurement
//--------------------------------------------------------------------
double seconds()
{
  struct timeval tm;
  double t;
  gettimeofday(&tm,NULL);
  t = (double) (tm.tv_sec) + ((double) (tm.tv_usec))/1.0e6;
  return t;
}


//====================================================================
int main(int argc, char **argv)
{
  Dbi* dbi;
  char sqlstmt[500];
  char hnr[20], pnr[20];
  int rc,i,j,k;
  long n, ncommit;
  double sec;
  string maxval;
  StringList result;

  // command line arguments
  string usage = (string)"Usage: " + argv[0] + " [-verbose] [-u <user>] "
    "[-p <passwd>] <dbname>\n";
  for (i=1; i<argc; i++) {
    if (0 == strcmp(argv[i], "-verbose")) {
      opt.verbose = 1;
    } else if (0 == strcmp(argv[i], "-u")) {
      i++;
      opt.user = argv[i];
    } else if (0 == strcmp(argv[i], "-p")) {
      i++;
      opt.password = argv[i];
    } else if ('-' == argv[i][0]) {
      fprintf(stderr, usage.c_str());
      return 1;
    } else {
      opt.dbname = argv[i];
    }
  }
  if (opt.dbname == "") {
      fprintf(stderr, usage.c_str());
      return 1;
  }

  // call constructor provided by linked in library
  dbi = ConstructDbi();

  // open DB
  rc = dbi->Login(opt.dbname.c_str(), opt.user.c_str(), opt.password.c_str());
  if (rc) {
    fprintf(stderr, "Can't open database: %s\n", dbi->LastError());
    return 2;
  }

  //
  // INSERT statements
  //
  sec = seconds();
  n = 0;
  ncommit = 0;

  if (dbi->Begin()) {
    fprintf(stderr, "SQL error: %s\n", dbi->LastError());
  }

  // loop over hersteller
  for (i=0; i<NUM_HERSTELLER; i++) {
    sprintf(hnr, "H%d", i);
    sprintf(sqlstmt, "INSERT INTO hersteller (hnr,name) VALUES "
            "('%s','Hersteller %d')", hnr, i);
    if (dbi->NonSelect(sqlstmt)) {
      fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
    }
    n++;
    if (0 == n % NUM_COMMIT) {
      dbi->Commit();
      ncommit++;
      dbi->Begin();
    }
    // produkte per hersteller
    for (j=0; j<NUM_PRODUKT; j++) {
      sprintf(pnr, "%sP%d", hnr, j);
      sprintf(sqlstmt, "INSERT INTO produkt (pnr,name,hnr,preis) VALUES "
              "('%s','Produkt %d','%s','%0.2f')", pnr, j, hnr, (float)(j+i)*0.3);
      n++;
      if (dbi->NonSelect(sqlstmt)) {
        fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
      }
      if (0 == n % NUM_COMMIT) {
        dbi->Commit();
        ncommit++;
        dbi->Begin();
      }
      // lieferungen per produkt
      for (k=0; k<NUM_LIEFERUNG; k++) {
        sprintf(sqlstmt, "INSERT INTO lieferung (lnr,pnr,datum,menge) VALUES "
                "('%ld','%s','2003%02d%02d','%0.2f')", n, pnr, 1+((i+j+k)%12),
                1+((i+j+k)%28), (float)(j+i+k)*0.2);
        n++;
        if (dbi->NonSelect(sqlstmt)) {
          fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
        }
        if (0 == n % NUM_COMMIT) {
          dbi->Commit();
          ncommit++;
          dbi->Begin();
        }
      }
    }
  }

  if (dbi->Commit()) {
    fprintf(stderr, "SQL error: %s\n", dbi->LastError());
  }
  ncommit++;
  printf("%ld Inserts (%ld Commits): %0.4f seconds\n",
         n, ncommit, seconds()-sec);

#if ! INSERT_ONLY

  //
  // SELECT without index
  //

  // statement that requires a sequential table scan
  sec = seconds();
  sprintf(sqlstmt, "SELECT max(menge) FROM lieferung");
  if (dbi->Select(sqlstmt, &result)) {
    fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  } else {
    maxval = *(result.begin());
    if (opt.verbose)
      printf("%s\n => %s\n", sqlstmt, maxval.c_str());
  }
  printf("Single table select: %0.4f seconds\n", seconds()-sec);

  // Join over three tables
  sec = seconds();
  sprintf(sqlstmt, "SELECT h.hnr,h.name FROM hersteller h, produkt p, "
          "lieferung l WHERE h.hnr=p.hnr AND p.pnr=l.pnr AND l.menge = '%s'",
          maxval.c_str());
  if (dbi->Select(sqlstmt, &result)) {
     fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  } else {
    if (opt.verbose) {
      StringList::iterator i;
      printf("%s\n => ", sqlstmt);
      for (i=result.begin(); i!=result.end(); i++)
        printf("%s|", i->c_str());
      printf("\n");
    }
  }
  printf("Join 3 tables (no index): %0.4f seconds\n", seconds()-sec);

  // Group over single table
  sec = seconds();
  sprintf(sqlstmt, "SELECT sum(menge),pnr FROM lieferung "
          "GROUP BY pnr ORDER BY 1 DESC");
  if (dbi->Select(sqlstmt, &result)) {
     fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  }
  printf("Group single table over non index field: %0.4f seconds\n",
         seconds()-sec);
  
  // Group over two tables
  sec = seconds();
  sprintf(sqlstmt, "SELECT sum(p.preis*l.menge),datum "
          "FROM lieferung l, produkt p WHERE l.pnr=p.pnr "
          "GROUP BY datum ORDER BY datum");
  if (dbi->Select(sqlstmt, &result)) {
     fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  } else {
    if (opt.verbose) {
      StringList::iterator i;
      printf("%s\n => ", sqlstmt);
      for (i=result.begin(); i!=result.end(); i++)
        printf("%s|", i->c_str());
      printf("\n");
    }
  }
  printf("Group two tables (no indices): %0.4f seconds\n", seconds()-sec);
  
  //
  // INDEX creation
  //
  sec = seconds();
  sprintf(sqlstmt, "CREATE INDEX l_pnr ON lieferung(pnr)");
  if (dbi->NonSelect(sqlstmt))
    fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  sprintf(sqlstmt, "CREATE INDEX p_hnr ON produkt(hnr)");
  if (dbi->NonSelect(sqlstmt))
    fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  // PostgreSQL requires analyze in addition
  dbi->NonSelect("ANALYZE lieferung");
  dbi->NonSelect("ANALYZE produkt");
  dbi->NonSelect("ANALYZE hersteller");
  printf("Create two indices: %0.4f seconds\n", seconds()-sec);

  //
  // SELECT with index
  //

  // Join over three tables
  sec = seconds();
  sprintf(sqlstmt, "SELECT h.hnr,h.name FROM hersteller h, produkt p, "
          "lieferung l WHERE h.hnr=p.hnr AND p.pnr=l.pnr AND l.menge = '%s'",
          maxval.c_str());
  if (dbi->Select(sqlstmt, &result)) {
     fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  } else {
    if (opt.verbose) {
      StringList::iterator i;
      printf("%s\n => ", sqlstmt);
      for (i=result.begin(); i!=result.end(); i++)
        printf("%s|", i->c_str());
      printf("\n");
    }
  }
  printf("Join 3 tables (with index): %0.4f seconds\n", seconds()-sec);

  // Group over single table
  sec = seconds();
  sprintf(sqlstmt, "SELECT sum(menge),pnr FROM lieferung "
          "GROUP BY pnr ORDER BY 1 DESC");
  if (dbi->Select(sqlstmt, &result)) {
     fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  } else {
    if (opt.verbose) {
      StringList::iterator i;
      printf("%s\n => ", sqlstmt);
      for (i=result.begin(); i!=result.end(); i++)
        printf("%s|", i->c_str());
      printf("\n");
    }
  }
  printf("Group single table over index field: %0.4f seconds\n",
         seconds()-sec);
  
  // Group over two tables
  sec = seconds();
  sprintf(sqlstmt, "SELECT sum(p.preis*l.menge),datum "
          "FROM lieferung l, produkt p WHERE l.pnr=p.pnr "
          "GROUP BY datum ORDER BY datum");
  if (dbi->Select(sqlstmt, &result)) {
     fprintf(stderr, "%s\nSQL error: %s\n", sqlstmt, dbi->LastError());
  } else {
    if (opt.verbose) {
      StringList::iterator i;
      printf("%s\n => ", sqlstmt);
      for (i=result.begin(); i!=result.end(); i++)
        printf("%s|", i->c_str());
      printf("\n");
    }
  }
  printf("Group two tables (with indices): %0.4f seconds\n", seconds()-sec);


#endif

  // close DB
  dbi->Logout();
  return 0;
}

